###
#Bass's Seen script.  v1.3.3  1/99 
#*!bass@*.psu.edu on Undernet.  Email: bass@shell.lazerlink.net
#This script maintains a database of *all* nicks seen online,
#not just those ppl who are in the userlist.  Simple flood 
#protection is built-in.  Wildcard searches are supported.
#
#Input is accepted from:
#       pub:  !seen <query> [#chan]
#       msg:  seen <query> [#chan]
#   and dcc:  .seen <query> [#chan]
#
#Queries can be in the following formats (public examples given):
#    'regular' !seen lamer; !seen lamest 			| SmartSearch-enabled query
#    'limited' !seennick lamer				| SmartSearch-bypassed query
#    'masked'  !seen *l?mer*; !seen *.lame.com; !seen *.edu #mychan
#
#Bonus feature:  !lastspoke <nick>  
#    You can use wildcard matching for <nick>.  The results for the
#    first match are returned.
###

#bs(limit) is the database record limit.  
set bs(limit) 3000

#bs(nicksize) is the maximum nickname length (9 on Undernet)
set bs(nicksize) 9

#bs(no_pub) is a list of channels you *don't* want the bot to post public 
#  replies to (public queries ignored).  Enter in lower case, eg: #lamer
set bs(no_pub) ""

#bs(quiet_chan) is a list of channels you want replies to requests sent  
#  to the person who made the query via notice. (The bot replies to  
#  public queries via notice.)  Enter in lower case.
set bs(quiet_chan) ""

#bs(no_log) is a list of channels you *don't* want the bot to log
#  data on.  Enter chans in lower case.
set bs(no_log) ""

#bs(log_only) is a list of channels you *only* want the bot to log
#  data on.  This is the opposite of bs(no_log).  Set it to "" if you 
#  want to log new channels the bot joins.  Enter chans in lower case.
set bs(log_only) ""

#bs(cmdchar) is what command character should be used for making public 
#  queries.  The default is "!".  Setting it to "" is a valid option.
set bs(cmdchar) "!"

#bs(flood) is used for flood protection, in the form x:y.  Any queries
#  beyond x in y seconds is considered a flood and ignored.
set bs(flood) 4:15
#bs(ignore) is used as a switch for ignoring flooders (1=on)
set bs(ignore) 1
#bs(ignore_time) is used to define the amount of time a flooder is 
#  ignored (minutes).  This is meaningless if bs(ignore) is 0.
set bs(ignore_time) 2

#bs(smartsearch) is a master enable/disable for SmartSearch.  SmartSearch ensures that
#  the most accurate and current results are returned for nick queries. (1=on)
set bs(smartsearch) 1

###### Don't edit below here, even if you do know Tcl ######
set bs(version) bseen1.3.3
if {[bind msg -|- help] != "*msg:help"} {bind msg -|- help *msg:help} ; #this is to fix a bind I didn't intend to use in a prev version (which screwed up msg'd help).  Sorry!

proc bs_filt {data} {
  regsub -all -- \\\\ $data \\\\\\\\ data
  regsub -all -- \\\[ $data \\\\\[ data
  regsub -all -- \\\] $data \\\\\] data
  regsub -all -- \\\} $data \\\\\} data
  regsub -all -- \\\{ $data \\\\\{ data
  regsub -all -- \\\" $data "" data
  return $data
}

proc bs_flood_init {} {
  global bs bs_flood_array
  if {![string match *:* $bs(flood)]} {putlog $bs(version): var bs(flood) not set correctly. ; return 1}
  set bs(flood_num) [lindex [split $bs(flood) :] 0]
  set bs(flood_time) [lindex [split $bs(flood) :] 1]
  set i [expr $bs(flood_num) - 1]
  while {$i >= 0} {
    set bs_flood_array($i) 0
    incr i -1
  }
}
bs_flood_init
proc bs_flood {nick uhost} {
  global bs bs_flood_array
  if {$bs(flood_num) == 0} {return 0}
  set i [expr $bs(flood_num) - 1]
  while {$i >= 1} {
    set bs_flood_array($i) $bs_flood_array([expr $i - 1])
    incr i -1
  }
  set bs_flood_array(0) [unixtime]
  if {[expr [unixtime] - $bs_flood_array([expr $bs(flood_num) - 1])] <= $bs(flood_time)} {
    putlog "$bs(version): Flood detected from $nick."
    if {$bs(ignore)} {newignore [join [maskhost *!*[string trimleft $uhost ~]]] $bs(version) flood $bs(ignore_time)}
    return  1
  } {return 0}
}

if {[lsearch -exact [bind time -|- "*2 * * * *"] bs_timedsave] > -1} {unbind time -|- "*2 * * * *" bs_timedsave} ; #backup frequency can be lower
bind time - "12 * * * *" bs_timedsave
proc bs_timedsave {min b c d e} {bs_save}

proc bs_save {} {
  global bs_list userfile
  if {[array size bs_list] == 0} {return 1}
  if {![string match */* $userfile]} {set name [lindex [split $userfile .] 0]} {
    set temp [split $userfile /] ; set temp [lindex $temp [expr [llength $temp]-1]]
    set name [lindex [split $temp .] 0]
  }
  if {[file exists bs_data.$name]} {
    catch {exec cp -f bs_data.$name bs_data.$name.bak}
  }
  set fd [open bs_data.$name w]
  set id [array startsearch bs_list]
  putlog "Backing up seen data..."
  while {[array anymore bs_list $id]} {
    set item [array nextelement bs_list $id]
    puts $fd "$bs_list($item)"
  }
  array donesearch bs_list $id
  close $fd
}

proc bs_read {} {
  global bs_list userfile
  if {![string match */* $userfile]} {set name [lindex [split $userfile .] 0]} {
    set temp [split $userfile /] ; set temp [lindex $temp [expr [llength $temp]-1]]
    set name [lindex [split $temp .] 0]
  }
  if {![file exists bs_data.$name]} {
    if {![file exists bs_data.$name.bak]} {
      putlog "Old seen data not found!"
      return 1
    } {
      exec cp bs_data.$name.bak bs_data.$name
      putlog "Old seen data not found! Using backup data."
    }
  }
  set fd [open bs_data.$name r]
  while {![eof $fd]} {
	set inp [gets $fd]
	if {[eof $fd]} {break}
	if {[string trim $inp " "] == ""} {continue}
      catch {
        set nick [lindex $inp 0]
	  set bs_list([string tolower $nick]) $inp
      }
  }
  close $fd
  putlog "Done loading [array size bs_list] seen records."
}

putlog "$bs(version):  -- Bass's SEEN loaded --"
if {![info exists bs_list] || [array size bs_list] == 0} {
  putlog "Loading seen database..."
  bs_read
}

bind part -|- * bs_part
proc bs_part {nick uhost hand chan} {bs_add $nick "[list $uhost] [unixtime] part $chan"}
bind join -|- * bs_join
proc bs_join {nick uhost hand chan} {bs_add $nick "[list $uhost] [unixtime] join $chan"}
bind sign -|- * bs_sign
proc bs_sign {nick uhost hand chan reason} {bs_add $nick "[list $uhost] [unixtime] quit $chan [split $reason]"}
bind kick -|- * bs_kick
proc bs_kick {nick uhost hand chan knick reason} {bs_add $knick "[getchanhost $knick $chan] [unixtime] kick $chan [list $nick] [list $reason]"}
bind nick -|- * bs_nick
proc bs_nick {nick uhost hand chan newnick} {bs_add $nick "[list $uhost] [unixtime] nick $chan [list $newnick]"}
bind splt -|- * bs_splt
proc bs_splt {nick uhost hand chan} {bs_add $nick "[list $uhost] [unixtime] splt $chan"}
bind rejn -|- * bs_rejn
proc bs_rejn {nick uhost hand chan} {bs_add $nick "[list $uhost] [unixtime] rejn $chan"}
bind chon -|- * bs_chon
proc bs_chon {hand idx} {
  foreach item [dcclist] {
    if {[lindex $item 3] != "CHAT"} {continue}
    if {[lindex $item 0] == $idx} {
      bs_add $hand "[lindex $item 2] [unixtime] chon" ; break
    }
  }
}
if {[lsearch -exact [bind chof -|- *] bs_chof] > -1} {unbind chof -|- * bs_chof} ; #this bind isn't needed any more
bind chjn -|- * bs_chjn
proc bs_chjn {bot hand channum flag sock from} {bs_add $hand "[string trimleft $from ~] [unixtime] chjn $bot"}
bind chpt -|- * bs_chpt
proc bs_chpt {bot hand args} {set old [bs_search ? [string tolower $hand]] ; if {$old != "0"} {bs_add $hand "[string trim [lindex $old 1] ()] [unixtime] chpt $bot"} }

bind dcc n|- unseen bs_unseen
proc bs_unseen {hand idx args} {
  global bs_list
  set tot 0 ; set chan [string tolower [lindex $args 0]]
  set id [array startsearch bs_list]
  while {[array anymore bs_list $id]} {
    set item [array nextelement bs_list $id]
    if {$chan == [string tolower [lindex $bs_list($item) 4]]} {incr tot ; lappend remlist $item}
  }
  array donesearch bs_list $id
  if {$tot > 0} {foreach item $remlist {unset bs_list($item)}}
  putlog "$hand removed $chan from the bseen database.  $tot entries removed."
  putidx $idx "$chan successfully removed.  $tot entries deleted from the bseen database."
}

bind dcc -|- seen bs_dccreq1
bind dcc -|- seennick bs_dccreq2
proc bs_dccreq1 {hand idx args} {bs_dccreq $hand $idx [bs_filt [join $args]] 0}
proc bs_dccreq2 {hand idx args} {bs_dccreq $hand $idx [bs_filt [join $args]] 1}
proc bs_dccreq {hand idx args no} {
  if {[string match *\\\** [lindex $args 0]]} {
    set output [bs_seenmask dcc $hand $args]
    if {$output == "No matches were found."} {putallbots "bs_botsearch 0 $idx $hand $args"}
    if {[string match "I'm not on *" $output]} {putallbots "bs_botsearch 0 $idx $hand $args"}
    putdcc $idx $output
    return 1
  }
  set args [lindex $args 0]
  set output [bs_output dcc $hand $args $no]
  if {$output == 0} {return 0}
  if {[lrange $output 1 4] == "I don't remember seeing"} {putallbots "bs_botsearch 0 $idx $hand $args"}
  putdcc $idx "$output"
  return 1
}

bind bot -|- bs_botsearch bs_botsearch
proc bs_botsearch {from cmd args} {
  global botnick
  set args [join $args]
  set command [lindex $args 0]
  set target [lindex $args 1]
  set nick [lindex $args 2]
  set search [lrange $args 3 e]
  if {[string match *\\\** $search]} {
    set output [bs_seenmask bot $nick [join $search]]
    if {$output != "No matches were found." && ![string match "I'm not on *" $output]} {putbot $from "bs_botsearch_reply $command \{$target\} {$nick, $botnick says:  [bs_filt $output]}"}
  } {
    set output [bs_output bot $nick [join $search] 0]
    if {$output != 0 && [lrange $output 1 4] != "I don't remember seeing"} {putbot $from "bs_botsearch_reply $command \{$target\} {$nick, $botnick says:  [bs_filt $output]}"}
  }
}
if {[info exists bs(bot_delay)]} {unset bs(bot_delay)}
bind bot -|- bs_botsearch_reply bs_botsearch_reply
proc bs_botsearch_reply {from cmd args} {
  global bs
  set args [join $args]
  if {[lindex [lindex $args 2] 5] == "not" || [lindex [lindex $args 2] 4] == "not"} {return 0}
  if {![info exists bs(bot_delay)]} {
    set bs(bot_delay) on
    utimer 10 {unset bs(bot_delay)} 
    if {![lindex $args 0]} {putdcc [lindex $args 1] "[join [lindex $args 2]]"} {
      puthelp "[lindex $args 1] :[join [lindex $args 2]]"
    }
  }
}

bind msg -|- seen bs_msgreq1
bind msg -|- seennick bs_msgreq2
proc bs_msgreq1 {nick uhost hand args} {bs_msgreq $nick $uhost $hand [bs_filt [join $args]] 0}
proc bs_msgreq2 {nick uhost hand args} {bs_msgreq $nick $uhost $hand [bs_filt [join $args]] 1}
proc bs_msgreq {nick uhost hand args no} { 
  if {[bs_flood $nick $uhost]} {return 0}
  if {[string match *\\\** [lindex $args 0]]} {
    set output [bs_seenmask msg $nick $args] 
    if {$output == "No matches were found."} {putallbots "bs_botsearch 1 \{notice $nick\} $nick $args"}
    if {[string match "I'm not on *" $output]} {putallbots "bs_botsearch 1 \{notice $nick\} $nick $args"}
    puthelp "notice $nick :$output" 
    return 1
  }
  set search [lindex $args 0]
  set output [bs_output $search $nick $search $no]
  if {$output == 0} {return 0}
  if {[lrange $output 1 4] == "I don't remember seeing"} {putallbots "bs_botsearch 1 \{notice $nick\} $nick $args"}
  puthelp "notice $nick :$output"
  return 1
}

bind pub -|- [string trim $bs(cmdchar)]seen bs_pubreq1
bind pub -|- [string trim $bs(cmdchar)]seennick bs_pubreq2
proc bs_pubreq1 {nick uhost hand chan args} {bs_pubreq $nick $uhost $hand $chan [bs_filt [join $args]] 0}
proc bs_pubreq2 {nick uhost hand chan args} {bs_pubreq $nick $uhost $hand $chan [bs_filt [join $args]] 1}
proc bs_pubreq {nick uhost hand chan args no} {
  global botnick bs
  if {[bs_flood $nick $uhost]} {return 0}
  set i 0 
  if {[lsearch -exact $bs(no_pub) [string tolower $chan]] >= 0} {return 0}
  if {$bs(log_only) != "" && [lsearch -exact $bs(log_only) [string tolower $chan]] == -1} {return 0}
  if {[lsearch -exact $bs(quiet_chan) [string tolower $chan]] >= 0} {set target "notice $nick"} {set target "privmsg $chan"}
  if {[string match *\\\** [lindex $args 0]]} {
    set output [bs_seenmask $chan $hand $args]
    if {$output == "No matches were found."} {putallbots "bs_botsearch 1 \{$target\} $nick $args"}
    if {[string match "I'm not on *" $output]} {putallbots "bs_botsearch 1 \{$target\} $nick $args"}
    puthelp "$target :$output"
    return 1
  }
  set data [string trimright [lindex $args 0] ?!.,]
  if {[string tolower $nick] == [string tolower $data] } { 
    puthelp "$target :$nick, go look in a mirror."
    return 1
  }
  if {[string tolower $data] == [string tolower $botnick] } { 
    puthelp "$target :$nick, I'm right here.  Quit wasting my time!"
    return 1
  }
  if {[onchan $data $chan]} { 
    puthelp "$target :$nick, $data is right here!" 
    return 1
  }
  set output [bs_output $chan $nick $data $no]
  if {$output == 0} {return 0}
  if {[lrange $output 1 4] == "I don't remember seeing"} {putallbots "bs_botsearch 1 \{$target\} $nick $args"}
  puthelp "$target :$output"
  return 1
}

proc bs_output {chan nick data no} {
  global botnick bs version bs_list
  set data [string tolower [string trimright [lindex [bs_filt $data] 0] ?!.,]]
  if {$data == ""} {return 0}
  if {[string tolower $nick] == [string tolower $data]} { 
    return "$nick, go look in a mirror."
  }
  if {[string tolower $data] == [string tolower $botnick]} { 
    return "$nick, I'm right here.  Quit wasting my time!"
  }
  if {[string length $data] > $bs(nicksize)} {return 0} 
  if {$bs(smartsearch) != 1} {set no 1}
  if {$no == 0} {
    set matches "" ; set hand "" ; set addy ""
    if {[lsearch -exact [array names bs_list] $data] != "-1"} { 
      set addy [lindex $bs_list([string tolower $data]) 1]
      set hand [finduser $addy]
      foreach item [bs_seenmask dcc ? [maskhost $addy]] {
        if {[lsearch -exact $matches $item] == -1} {lappend matches $item}
      }
    }
    if {[validuser $data]} {set hand $data}
    if {$hand != "*" && $hand != ""} {
      if {[string trimleft [lindex $version 1] 0]>1030000} {set hosts [getuser $hand hosts]} {set hosts [gethosts $hand]}
      foreach addr $hosts {
        foreach item [string tolower [bs_seenmask dcc ? $addr]] {
          if {[lsearch -exact [string tolower $matches] [string tolower $item]] == -1} {lappend matches $item}
        }
      }
    }
    if {$matches != ""} {
      set matches [join [join $matches]]
      set len [llength $matches]
      if {$len == 1} {return "[bs_search $chan $matches]"}
      if {$len > 99} {return "I found $len matches to your query; please refine it to see any output."}
      set matches [bs_sort $matches]
      if {[string tolower [lindex $matches 0]] == [string tolower $data]} {return [bs_search $chan [lindex $matches 0]]}
      if {$len < 6} {
        set output "I found $len matches to your query (sorted): [join $matches]."
        set output "$output  [bs_search $chan [lindex $matches 0]]"
        return $output
      } {
        set output "I found $len matches to your query.  Here are the 5 most recent (sorted): [join [lrange $matches 0 4]]."
        set output "$output  [bs_search $chan [lindex $matches 0]]"
        return $output
      }
    }
  }
  set temp [bs_search $chan $data]
  if {$temp != 0} { return "$temp" } {
    #$data not found in $bs_list, so search userfile
    if {![validuser [bs_filt $data]] || [string trimleft [lindex $version 1] 0]<1030000} { 
      return "$nick, I don't remember seeing \"$data\""
    } {
      if {[getuser $data laston] != ""} {
        set seen [getuser $data laston]
        if {($chan != [lindex $seen 1] || $chan == "bot" || $chan == "msg" || $chan == "dcc") && [validchan [lindex $seen 1]] && [lindex [channel info [lindex $seen 1]] 23] == "+secret"} {
          set chan "-secret-"
        } {
          set chan [lindex $seen 1]
        }
        return "$data was last seen on $chan [bs_when [lindex $seen 0]] ago."
      }
    }
  }
}

proc bs_search {chan nick} {
  global bs_list
  if {![info exists bs_list]} {return 0}
  if {[lsearch -exact [array names bs_list] [string tolower $nick]] != "-1"} { 
    set data $bs_list([string tolower $nick])
    set nick [lindex $data 0]
    set addy [lindex $data 1]
    set time [lindex $data 2]
    if {([string tolower $chan] != [string tolower [lindex $data 4]] || $chan == "dcc" || $chan == "msg" || $chan == "bot") && [validchan [lindex $data 4]] && [lindex [channel info [lindex $data 4]] 23] == "+secret"} {
      set chan "-secret-"
    } {
      set chan [lindex $data 4]
    }
    set marker 0
    switch -- [lindex $data 3] {
      part { set output "$nick ($addy) was last seen parting $chan [bs_when $time] ago." }
      quit { set output "$nick ($addy) was last seen quitting from $chan [bs_when $time] ago stating ([join [lrange $data 5 e]])." }
      kick { set output "$nick ($addy) was last seen being kicked from $chan by [lindex $data 5] [bs_when $time] ago with the reason ([join [lrange $data 6 e]])." }
      nick { 
        set output "$nick ($addy) was last seen changing nicks to [lindex $data 5] on [lindex $data 4] [bs_when $time] ago." 
        if {[validchan [lindex $data 4]]} {
          if {[onchan [lindex $data 5] [lindex $data 4]]} {lappend output [lindex $data 5] is still there.} {lappend output I don't see [lindex $data 5] now, though.]}
        }
      }
      splt { set output "$nick ($addy) was last seen parting $chan due to a split [bs_when $time] ago." }
      rejn { 
        set output "$nick ($addy) was last seen rejoining $chan from a split [bs_when $time] ago." 
        if {[validchan $chan]} {
          if {[onchan $nick $chan]} {lappend output $nick is still on $chan.} {lappend output I don't see $nick on $chan now, though.}
        }
      }
      join { 
        set output "$nick ($addy) was last seen joining $chan [bs_when $time] ago." 
        if {[validchan $chan]} {
          if {[onchan $nick $chan]} {lappend output $nick is still on $chan.} {lappend output I don't see $nick on $chan now, though.}
        }
      }
      chon { 
        set output "$nick ($addy) was last seen joining the partyline [bs_when $time] ago." 
        set lnick [string tolower $nick]
        foreach item [whom *] {
          if {$lnick == [string tolower [lindex $item 0]]} {
            lappend output $nick is on the partyline right now.
            set marker 1
            break
          } 
        }
        if {$marker == 0} {lappend output I don't see $nick on the partyline now, though.}
      }
      chof { 
        set output "$nick ($addy) was last seen leaving the partyline [bs_when $time] ago." 
        set lnick [string tolower $nick]
        foreach item [whom *] {
          if {$lnick == [string tolower [lindex $item 0]]} {
            lappend output $nick is on the partyline in [lindex $item 1] still.
            break
          } 
        }
      }
      chjn { 
        set output "$nick ($addy) was last seen joining the partyline on $chan [bs_when $time] ago." 
        set lnick [string tolower $nick]
        foreach item [whom *] {
          if {$lnick == [string tolower [lindex $item 0]]} {
            lappend output $nick is on the partyline right now.
            set marker 1
            break
          } 
        }
        if {$marker == 0} {lappend output I don't see $nick on the partyline now, though.}
      }
      chpt { 
        set output "$nick ($addy) was last seen leaving the partyline from $chan [bs_when $time] ago." 
        set lnick [string tolower $nick]
        foreach item [whom *] {
          if {$lnick == [string tolower [lindex $item 0]]} {
            lappend output $nick is on the partyline in [lindex $item 1] still.
            break
          } 
        }
      }

      default {set output "error"}
    }
    return $output
  } {
    return 0
  }
}

proc bs_when {lasttime} {
  #This is equiv to mIRC's $duration() function
  set years 0 
  set days 0 
  set hours 0 
  set mins 0
  set time [expr [unixtime] - $lasttime]
  if {$time < 60} {
    return "only $time seconds"
  }
  if {$time >= 31536000} {
    set years [expr int([expr $time/31536000])]
    set time [expr $time - [expr 31536000*$years]]
  }
  if {$time >= 86400} {
    set days [expr int([expr $time/86400])]
    set time [expr $time - [expr 86400*$days]]
  }
  if {$time >= 3600} {
    set hours [expr int([expr $time/3600])]
    set time [expr $time - [expr 3600*$hours]]
  }
  if {$time >= 60} {set mins [expr int([expr $time/60])]}
  if {$years == 0} {set output ""} elseif {$years == 1} {set output "1 year,"} {set output "$years years,"}
  if {$days == 1} {lappend output "1 day,"} elseif {$days > 1} {lappend output "$days days,"}
  if {$hours == 1} {lappend output "1 hour,"} elseif {$hours > 1} {lappend output "$hours hours,"}
  if {$mins == 1} {lappend output "1 minute"} elseif {$mins > 1} {lappend output "$mins minutes"}
  return [string trimright [join $output] ", "]
}

proc bs_add {nick data} {
  global bs_list bs
  if {[lsearch -exact $bs(no_log) [string tolower [lindex $data 3]]] >= 0} {return 1}
  if {$bs(log_only) != "" && [lsearch -exact $bs(log_only) [string tolower [lindex $data 3]]] == -1} {return 0}
  set i 0 
  set temp ""
  while {$i < [string length $nick]} {
    if {[string index $nick $i] == "\\"} {append temp "\\\\"} elseif {
    [string index $nick $i] == "\{"} {append temp "\\\{"} elseif {
    [string index $nick $i] == "\}"} {append temp "\\\}"} {
    append temp [string index $nick $i]}
    incr i
  }
  set nick2 $temp
  if {[lindex $nick2 0] != $nick} {
    set nick2 [list $nick2] 
  }
  if {[info exists bs_list] && [lsearch -exact [array names bs_list] [string tolower $nick]] > -1} {
    unset bs_list([string tolower $nick])
    set bs_list([string tolower $nick]) "$nick2 $data"
  } {
    set bs_list([string tolower $nick]) "$nick2 $data"
    while {[array size bs_list] > $bs(limit)} {
      set id [array startsearch bs_list]
      set record [unixtime]
      array set name {0 "" 1 "" 2 "" 3 "" 4 "" 5 "" 6 "" 7 "" 8 "" 9 ""}
      while {[array anymore bs_list $id]} {
        set item [array nextelement bs_list $id]
        catch {
          if {[lindex $bs_list($item) 2] <= $record} {
            set record [lindex $bs_list($item) 2]
            set name(9) $name(8)
            set name(8) $name(7)
            set name(7) $name(6)
            set name(6) $name(5)
            set name(5) $name(4)
            set name(4) $name(3)
            set name(3) $name(2)
            set name(2) $name(1)
            set name(1) $name(0)
            set name(0) $item
          }
        }
      }
      array donesearch bs_list $id
      unset bs_list($name(0)) bs_list($name(1)) bs_list($name(2)) bs_list($name(3)) bs_list($name(4)) bs_list($name(5)) bs_list($name(6)) bs_list($name(7)) bs_list($name(8)) bs_list($name(9)) 
    }
  }
}

proc bs_seenmask {ch nick args} {
  global bs_list
  set matches "" ; set temp "" ; set i 0
  set args [bs_filt [join [join $args]]]
  set chan [lindex $args 1]
  if {$chan != "" && [string trimleft $chan #] != $chan} {
    if {![validchan $chan]} {return "I'm not on $chan."} {set chan [string tolower $chan]}
  } { set $chan "" }
  if {![info exists bs_list]} {return "No matches were found."}
  set data [string tolower [lindex $args 0]]
  set id [array startsearch bs_list]
  while {[array anymore bs_list $id]} {
    set item [array nextelement bs_list $id]
    if {$item == ""} {continue}
    set i 0 
    set temp ""
    set match [lindex $bs_list($item) 0]
    set addy [lindex $bs_list($item) 1]
    #Sorry about using 'catch', but some nicks in the list were causing TCL errors.
    catch {
      if {[string match $data $item![string tolower $addy]]} {
        if {[string match *\\\\* $match] || [string match *\\\{* $match] || [string match *\\\}* $match]} {
          while {$i < [string length $match]} {
            if {[string index $match $i] == "\\"} {append temp "\\\\"} elseif {
              [string index $match $i] == "\}"} {append temp "\\\}"} elseif {
              [string index $match $i] == "\{"} {append temp "\\\{"} {
              append temp [string index $match $i]
            }
            incr i
          }
          set match [list $temp]
        }
        if {$chan != ""} {
          if {[string match $chan [string tolower [lindex $bs_list($item) 4]]]} {
            set matches "$matches $match"
          }
        } {
            set matches "$matches $match"
        }
      }
    }
  }
  array donesearch bs_list $id
  set matches [string trim $matches " "]
  if {$nick == "?"} {return $matches}
  set len [llength $matches]
  if {$len == 0} {return "No matches were found."}
  if {$len == 1} {return "[bs_output $ch $nick [join [join $matches]] 1]"}
  if {$len > 99} {return "I found $len matches to your query; please refine it to see any output."}
  if {$len < 15} {
    set matches [join [bs_sort $matches]]
    set output "I found $len matches to your query (sorted): $matches."
    if {[llength $matches]<5} {set output "$output  [bs_output $ch $nick [lindex [bs_filt $matches] 0] 1]"}
    return $output
  } {
    return "I found $len matches to your query.  Here are the 15 most recent (sorted): [join [lrange [bs_sort $matches] 0 14]]."
  }
} 

proc bs_sort {data} {
  global bs_list
  set len [expr [llength $data] - 1]
  set i $len
  while {$i>0} {
    set j $len
    while {$j>[expr $len - $i]} {
      set item1 [join [string tolower [lindex $data $j]]]
      set item2 [join [string tolower [lindex $data [expr $j - 1]]]]
      if {[lindex $bs_list($item1) 2] > [lindex $bs_list($item2) 2]} {
        set data [lreplace $data [expr $j - 1] $j [lindex $data $j] [lindex $data [expr $j - 1]]]
      }
      incr j -1
    }
    incr i -1
  }
  return [join $data]
}

bind dcc -|- seenstats bs_dccstats
proc bs_dccstats {hand idx args} {putdcc $idx "[bs_stats]"; return 1}
bind pub -|- [string trim $bs(cmdchar)]seenstats bs_pubstats
proc bs_pubstats {nick uhost hand chan args} {
  global bs
  if {[bs_flood $nick $uhost]} {return 0}
  if {[lsearch -exact $bs(no_pub) [string tolower $chan]] >= 0} {return 0}
  if {$bs(log_only) != "" && [lsearch -exact $bs(log_only) [string tolower $chan]] == -1} {return 0}
  if {[lsearch -exact $bs(quiet_chan) [string tolower $chan]] >= 0} {set target "notice $nick"} {set target "privmsg $chan"}
  puthelp "$target :[bs_stats]" ; return 1
}
bind msg -|- seenstats bs_msgstats
proc bs_msgstats {nick uhost hand args} { 
  if {[bs_flood $nick $uhost]} {return 0}
  puthelp "notice $nick :[bs_stats]" ; return 1
}

proc bs_stats {} {
  global bs_list bs
  set id [array startsearch bs_list]
  set bs_record [unixtime]
  set totalm 0 ; set temp ""
  while {[array anymore bs_list $id]} {
    set item [array nextelement bs_list $id]
    catch {
      set tok [lindex $bs_list($item) 2]
      if {$tok == ""} {putlog "Damaged seen record: $item" ; continue}
      if {[lindex $bs_list($item) 2] < $bs_record} {
        set bs_record [lindex $bs_list($item) 2]
        set name $item
      }
      set addy [string tolower [maskhost [lindex $bs_list($item) 1]]]
      if {[lsearch -exact $temp $addy] == -1} {incr totalm ; lappend temp $addy}
    }
  }
  array donesearch bs_list $id
  return "Currently I am tracking [array size bs_list]/$bs(limit) nicks, which comprise $totalm unique uhosts.  The oldest record is [lindex $bs_list($name) 0]'s, which is from [bs_when $bs_record] ago."
}

bind dcc -|- chanstats bs_dccchanstats
proc bs_dccchanstats {hand idx args} {putdcc $idx "[bs_chanstats [lindex $args 0]]"; return 1}
bind pub -|- [string trim $bs(cmdchar)]chanstats bs_pubchanstats
proc bs_pubchanstats {nick uhost hand chan args} {
  global bs
  set chan [string tolower $chan]
  if {[bs_flood $nick $uhost]} {return 0}
  if {[lsearch -exact $bs(no_pub) $chan] >= 0} {return 0}
  if {$bs(log_only) != "" && [lsearch -exact $bs(log_only) [string tolower $chan]] == -1} {return 0}
  if {[lsearch -exact $bs(quiet_chan) $chan] >= 0} {set target "notice $nick"} {set target "privmsg $chan"}
  if {[lindex $args 0] != ""} {set chan [lindex $args 0]}
  puthelp "$target :[bs_chanstats $chan]" ; return 1
}
bind msg -|- chanstats bs_msgchanstats
proc bs_msgchanstats {nick uhost hand args} { 
  if {[bs_flood $nick $uhost]} {return 0}
  puthelp "notice $nick :[bs_chanstats [lindex $args 0]]" ; return 1
}
proc bs_chanstats {chan} {
  global bs_list
  set chan [string tolower $chan]
  if {![validchan $chan]} {return "I'm not on $chan."}
  set id [array startsearch bs_list]
  set bs_record [unixtime] ; set totalc 0 ; set totalm 0 ; set temp ""
  while {[array anymore bs_list $id]} {
    set item [array nextelement bs_list $id]
    catch {
      set time [lindex $bs_list($item) 2]
      if {$time == ""} {continue}
      if {$chan == [string tolower [lindex $bs_list($item) 4]]} {
        if {$time < $bs_record} {
          set bs_record $time
          #set name $item
        }
        incr totalc
        set addy [string tolower [maskhost [lindex $bs_list($item) 1]]]
        if {[lsearch -exact $temp $addy] == -1} {incr totalm ; lappend temp $addy}
      }
    }
  }
  array donesearch bs_list $id
  set total [array size bs_list]
  return "$chan is the source of [expr 100*$totalc/$total]% ($totalc/$total) of the seen database entries.  In $chan, there were a total of $totalm unique uhosts seen in the last [bs_when $bs_record]."
}

bind pub -|- [string trim $bs(cmdchar)]lastspoke lastspoke
proc lastspoke {nick uhost hand chan args} {
  global bs botnick
  if {[bs_flood $nick $uhost]} {return 0}
  if {[lsearch -exact $bs(no_pub) [string tolower $chan]] >= 0} {return 0}
  if {[lsearch -exact $bs(quiet_chan) [string tolower $chan]] >= 0} {set target "notice $nick"} {set target "privmsg $chan"}
  if {$bs(log_only) != "" && [lsearch -exact $bs(log_only) [string tolower $chan]] == -1} {return 0}
  set data [lindex $args 0]
  if {[string match *\** $data]} {
    set i 0
    while {$i < [string length $args]} {
      if {[string index $args $i] == "\["} {append temp \\\[} elseif {
        [string index $args $i] == "\]"} {append temp \\\]} elseif {
        [string index $args $i] == "\\"} {append temp \\\\} {
        append temp [string index $args $i]
      }
      incr i
    }
    set data [lindex [string tolower $temp] 0]
    set chanlist [string tolower [chanlist $chan]]
    if {[lsearch -glob $chanlist [string tolower $data]] > -1} {set data [lindex [chanlist $chan] [lsearch -glob $chanlist $data]]}
  }
  if {[onchan $data $chan]} { 
    if {[string tolower $data] == [string tolower $botnick]} {puthelp "$target :$nick, must you waste my time?" ; return 1}
    set time [getchanidle $data $chan]
    if {$time >= 60} {
      set hours [expr int([expr $time / 60])]
      set time [expr $time - ($hours * 60)]
      if {$hours > 1} {set output "$hours hours, $time minutes"} {set output "1 hour, $time minutes"}
    } {set output "$time minutes"}
    puthelp "$target :$data last uttered a word on $chan $output ago." 
  }
  return 1
} 
bind msgm -|- "help seen" bs_help_msg_seen
bind msgm -|- "help chanstats" bs_help_msg_chanstats
bind msgm -|- "help seenstats" bs_help_msg_seenstats
proc bs_help_msg_seen {nick uhost hand args} {
  if {[bs_flood $nick $uhost]} {return 0}
  puthelp "notice $nick :###  seen <query> \[chan\]"
  puthelp "notice $nick :   Queries can be in the following formats:"
  puthelp "notice $nick :     'regular':  seen lamer; seen lamest "
  puthelp "notice $nick :     'masked':   seen *l?mer*; seen *.lame.com; seen *.edu #mychan"
  return 0
}
proc bs_help_msg_chanstats {nick uhost hand args} {
  if {[bs_flood $nick $uhost]} {return 0}
  puthelp "notice $nick :###  chanstats <chan>"
  puthelp "notice $nick :   Returns the usage statistics of #chan in the seen database."
  return 0
}
proc bs_help_msg_seenstats {nick uhost hand args} {
  if {[bs_flood $nick $uhost]} {return 0}
  puthelp "notice $nick :###  seenstats"
  puthelp "notice $nick :   Returns the status of the bseen database."
  return 0
}
bind dcc -|- help bs_help_dcc
proc bs_help_dcc {hand idx args} {
  switch -- $args {
    seen {
      putidx $idx "###  Bass's Seen script v1.3.3.  *!bass@*.psu.edu on Undernet."
      putidx $idx "###  seen <query> \[chan\]"
      putidx $idx "   Queries can be in the following formats:"
      putidx $idx "     'regular':  seen lamer; seen lamest "
      putidx $idx "     'masked':   seen *l?mer*; seen *.lame.com; seen *.edu #mychan"
    }
    chanstats {
      putidx $idx "###  Bass's Seen script v1.3.3.  *!bass@*.psu.edu on Undernet."
      putidx $idx "###  chanstats <chan>"
      putidx $idx "   Returns the usage statistics of #chan in the seen database."
    }
    seenstats {
      putidx $idx "###  Bass's Seen script v1.3.3.  *!bass@*.psu.edu on Undernet."
      putidx $idx "###  seenstats"
      putidx $idx "   Returns the status of the bseen database."
    }
    unseen {
      if {[matchattr $hand n]} {
        putidx $idx "###  Bass's Seen script v1.3.3.  *!bass@*.psu.edu on Undernet."
        putidx $idx "###  unseen <chan>"
        putidx $idx "   Deletes all <chan> entries from the bseen database."
      }
    }
    default {*dcc:help $hand $idx [join $args] ; return 0} 
 }
  return 1
}
